@@ -1,8 +1,10 @@
 /*
  * Copyright (C) 2010 Karel Zak <kzak@redhat.com>
  * Copyright (C) 2010 Davidlohr Bueso <dave@gnu.org>
+ *
+ * No copyright is claimed.  This code is in the public domain; do with
+ * it what you wish.
  */
-
 #include <stdio.h>
 #include <stdlib.h>
 #include <inttypes.h>
@@ -10,12 +12,14 @@
 #include <errno.h>
 #include <sys/stat.h>
 #include <string.h>
+#include <strings.h>
 #include <assert.h>
 
 #include "c.h"
 #include "nls.h"
 #include "strutils.h"
 #include "bitops.h"
+#include "pathnames.h"
 
 static int STRTOXX_EXIT_CODE = EXIT_FAILURE;
 
@@ -170,13 +174,23 @@ check_suffix:
 
 		/* maximal divisor for last digit (e.g. for 0.05 is
 		 * frac_div=100, for 0.054 is frac_div=1000, etc.)
+		 *
+		 * Reduce frac if too large.
 		 */
-		while (frac_div < frac)
-			frac_div *= 10;
+		while (frac_div < frac) {
+			if (frac_div <= UINTMAX_MAX/10)
+				frac_div *= 10;
+			else
+				frac /= 10;
+		}
 
 		/* 'frac' is without zeros (5 means 0.5 as well as 0.05) */
-		for (i = 0; i < frac_zeros; i++)
-			frac_div *= 10;
+		for (i = 0; i < frac_zeros; i++) {
+			if (frac_div <= UINTMAX_MAX/10)
+				frac_div *= 10;
+			else
+				frac /= 10;
+		}
 
 		/*
 		 * Go backwardly from last digit and add to result what the
@@ -192,7 +206,7 @@ check_suffix:
 			frac /= 10;	/* remove last digit from frac */
 			frac_poz *= 10;
 
-			if (seg)
+			if (seg && seg_div / seg)
 				x += frac_base / (seg_div / seg);
 		} while (frac);
 	}
@@ -252,7 +266,9 @@ int parse_switch(const char *arg, const char *errmesg, ...)
 		if (strcmp(arg, a) == 0) {
 			va_end(ap);
 			return 1;
-		} else if (strcmp(arg, b) == 0) {
+		}
+
+		if (strcmp(arg, b) == 0) {
 			va_end(ap);
 			return 0;
 		}
@@ -304,123 +320,119 @@ char *strndup(const char *s, size_t n)
 }
 #endif
 
-static uint32_t _strtou32_or_err(const char *str, const char *errmesg, int base);
-static uint64_t _strtou64_or_err(const char *str, const char *errmesg, int base);
-
-int16_t strtos16_or_err(const char *str, const char *errmesg)
+/*
+ * convert strings to numbers; returns <0 on error, and 0 on success
+ */
+int ul_strtos64(const char *str, int64_t *num, int base)
 {
-	int32_t num = strtos32_or_err(str, errmesg);
-
-	if (num < INT16_MIN || num > INT16_MAX) {
-		errno = ERANGE;
-		err(STRTOXX_EXIT_CODE, "%s: '%s'", errmesg, str);
-	}
-	return num;
-}
-
-static uint16_t _strtou16_or_err(const char *str, const char *errmesg, int base)
-{
-	uint32_t num = _strtou32_or_err(str, errmesg, base);
-
-	if (num > UINT16_MAX) {
-		errno = ERANGE;
-		err(STRTOXX_EXIT_CODE, "%s: '%s'", errmesg, str);
-	}
-	return num;
-}
-
-uint16_t strtou16_or_err(const char *str, const char *errmesg)
-{
-	return _strtou16_or_err(str, errmesg, 10);
-}
-
-uint16_t strtox16_or_err(const char *str, const char *errmesg)
-{
-	return _strtou16_or_err(str, errmesg, 16);
-}
-
-int32_t strtos32_or_err(const char *str, const char *errmesg)
-{
-	int64_t num = strtos64_or_err(str, errmesg);
-
-	if (num < INT32_MIN || num > INT32_MAX) {
-		errno = ERANGE;
-		err(STRTOXX_EXIT_CODE, "%s: '%s'", errmesg, str);
-	}
-	return num;
-}
-
-static uint32_t _strtou32_or_err(const char *str, const char *errmesg, int base)
-{
-	uint64_t num = _strtou64_or_err(str, errmesg, base);
-
-	if (num > UINT32_MAX) {
-		errno = ERANGE;
-		err(STRTOXX_EXIT_CODE, "%s: '%s'", errmesg, str);
-	}
-	return num;
-}
-
-uint32_t strtou32_or_err(const char *str, const char *errmesg)
-{
-	return _strtou32_or_err(str, errmesg, 10);
-}
-
-uint32_t strtox32_or_err(const char *str, const char *errmesg)
-{
-	return _strtou32_or_err(str, errmesg, 16);
-}
-
-int64_t strtos64_or_err(const char *str, const char *errmesg)
-{
-	int64_t num;
 	char *end = NULL;
 
-	errno = 0;
 	if (str == NULL || *str == '\0')
-		goto err;
-	num = strtoimax(str, &end, 10);
+		return -(errno = EINVAL);
 
-	if (errno || str == end || (end && *end))
-		goto err;
+	errno = 0;
+	*num = (int64_t) strtoimax(str, &end, base);
 
-	return num;
-err:
-	if (errno == ERANGE)
-		err(STRTOXX_EXIT_CODE, "%s: '%s'", errmesg, str);
-
-	errx(STRTOXX_EXIT_CODE, "%s: '%s'", errmesg, str);
+	if (errno != 0)
+		return -errno;
+	if (str == end || (end && *end))
+		return -(errno = EINVAL);
+	return 0;
 }
 
-static uint64_t _strtou64_or_err(const char *str, const char *errmesg, int base)
+int ul_strtou64(const char *str, uint64_t *num, int base)
 {
-	uintmax_t num;
 	char *end = NULL;
+	int64_t tmp;
 
-	errno = 0;
 	if (str == NULL || *str == '\0')
-		goto err;
-	num = strtoumax(str, &end, base);
+		return -(errno = EINVAL);
 
-	if (errno || str == end || (end && *end))
-		goto err;
+	/* we need to ignore negative numbers, note that for invalid negative
+	 * number strtoimax() returns negative number too, so we do not
+	 * need to check errno here */
+	errno = 0;
+	tmp = (int64_t) strtoimax(str, &end, base);
+	if (tmp < 0)
+		errno = ERANGE;
+	else {
+		errno = 0;
+		*num = strtoumax(str, &end, base);
+	}
 
+	if (errno != 0)
+		return -errno;
+	if (str == end || (end && *end))
+		return -(errno = EINVAL);
+	return 0;
+}
+
+int ul_strtos32(const char *str, int32_t *num, int base)
+{
+	int64_t tmp;
+	int rc;
+
+	rc = ul_strtos64(str, &tmp, base);
+	if (rc == 0 && (tmp < INT32_MIN || tmp > INT32_MAX))
+		rc = -(errno = ERANGE);
+	if (rc == 0)
+		*num = (int32_t) tmp;
+	return rc;
+}
+
+int ul_strtou32(const char *str, uint32_t *num, int base)
+{
+	uint64_t tmp;
+	int rc;
+
+	rc = ul_strtou64(str, &tmp, base);
+	if (rc == 0 && tmp > UINT32_MAX)
+		rc = -(errno = ERANGE);
+	if (rc == 0)
+		*num = (uint32_t) tmp;
+	return rc;
+}
+
+/*
+ * Convert strings to numbers in defined range and print message on error.
+ *
+ * These functions are used when we read input from users (getopt() etc.). It's
+ * better to consolidate the code and keep it all based on 64-bit numbers then
+ * implement it for 32 and 16-bit numbers too.
+ */
+int64_t str2num_or_err(const char *str, int base, const char *errmesg,
+			     int64_t low, int64_t up)
+{
+	int64_t num = 0;
+	int rc;
+
+	rc = ul_strtos64(str, &num, base);
+	if (rc == 0 && ((low && num < low) || (up && num > up)))
+		rc = -(errno = ERANGE);
+
+	if (rc) {
+		if (errno == ERANGE)
+			err(STRTOXX_EXIT_CODE, "%s: '%s'", errmesg, str);
+		errx(STRTOXX_EXIT_CODE, "%s: '%s'", errmesg, str);
+	}
 	return num;
-err:
-	if (errno == ERANGE)
-		err(STRTOXX_EXIT_CODE, "%s: '%s'", errmesg, str);
-
-	errx(STRTOXX_EXIT_CODE, "%s: '%s'", errmesg, str);
 }
 
-uint64_t strtou64_or_err(const char *str, const char *errmesg)
+uint64_t str2unum_or_err(const char *str, int base, const char *errmesg, uint64_t up)
 {
-	return _strtou64_or_err(str, errmesg, 10);
-}
+	uint64_t num = 0;
+	int rc;
 
-uint64_t strtox64_or_err(const char *str, const char *errmesg)
-{
-	return _strtou64_or_err(str, errmesg, 16);
+	rc = ul_strtou64(str, &num, base);
+	if (rc == 0 && (up && num > up))
+		rc = -(errno = ERANGE);
+
+	if (rc) {
+		if (errno == ERANGE)
+			err(STRTOXX_EXIT_CODE, "%s: '%s'", errmesg, str);
+		errx(STRTOXX_EXIT_CODE, "%s: '%s'", errmesg, str);
+	}
+	return num;
 }
 
 double strtod_or_err(const char *str, const char *errmesg)
@@ -444,36 +456,15 @@ err:
 	errx(STRTOXX_EXIT_CODE, "%s: '%s'", errmesg, str);
 }
 
-long strtol_or_err(const char *str, const char *errmesg)
+long double strtold_or_err(const char *str, const char *errmesg)
 {
-	long num;
+	double num;
 	char *end = NULL;
 
 	errno = 0;
 	if (str == NULL || *str == '\0')
 		goto err;
-	num = strtol(str, &end, 10);
-
-	if (errno || str == end || (end && *end))
-		goto err;
-
-	return num;
-err:
-	if (errno == ERANGE)
-		err(STRTOXX_EXIT_CODE, "%s: '%s'", errmesg, str);
-
-	errx(STRTOXX_EXIT_CODE, "%s: '%s'", errmesg, str);
-}
-
-unsigned long strtoul_or_err(const char *str, const char *errmesg)
-{
-	unsigned long num;
-	char *end = NULL;
-
-	errno = 0;
-	if (str == NULL || *str == '\0')
-		goto err;
-	num = strtoul(str, &end, 10);
+	num = strtold(str, &end);
 
 	if (errno || str == end || (end && *end))
 		goto err;
@@ -502,11 +493,28 @@ uintmax_t strtosize_or_err(const char *str, const char *errmesg)
 
 void strtotimeval_or_err(const char *str, struct timeval *tv, const char *errmesg)
 {
-	double user_input;
+	long double user_input;
 
-	user_input = strtod_or_err(str, errmesg);
+	user_input = strtold_or_err(str, errmesg);
 	tv->tv_sec = (time_t) user_input;
-	tv->tv_usec = (long)((user_input - tv->tv_sec) * 1000000);
+	tv->tv_usec = (suseconds_t)((user_input - tv->tv_sec) * 1000000);
+}
+
+void strtotimespec_or_err(const char *str, struct timespec *ts, const char *errmesg)
+{
+	long double user_input;
+
+	user_input = strtold_or_err(str, errmesg);
+	ts->tv_sec = (time_t) user_input;
+	ts->tv_nsec = (long)((user_input - ts->tv_sec) * 1000000000);
+}
+
+time_t strtotime_or_err(const char *str, const char *errmesg)
+{
+	int64_t user_input;
+
+	user_input = strtos64_or_err(str, errmesg);
+	return (time_t) user_input;
 }
 
 /*
@@ -599,24 +607,44 @@ char *size_to_human_string(int options, uint64_t bytes)
 
 	/* round */
 	if (frac) {
+		/* get 3 digits after decimal point */
+		if (frac >= UINT64_MAX / 1000)
+			frac = ((frac / 1024) * 1000) / (1ULL << (exp - 10)) ;
+		else
+			frac = (frac * 1000) / (1ULL << (exp)) ;
+
 		if (options & SIZE_DECIMAL_2DIGITS) {
-			frac = (frac / (1ULL << (exp - 10)) + 5) / 10;
-			if (frac % 10 == 0)
-				frac /= 10;	/* convert N.90 to N.9 */
+			/* round 4/5 and keep 2 digits after decimal point */
+			frac = (frac + 5) / 10 ;
 		} else {
-			frac = (frac / (1ULL << (exp - 10)) + 50) / 100;
-			if (frac == 10)
-				dec++, frac = 0;
+			/* round 4/5 and keep 1 digit after decimal point */
+			frac = ((frac + 50) / 100) * 10 ;
+		}
+
+		/* rounding could have overflowed */
+		if (frac == 100) {
+			dec++;
+			frac = 0;
 		}
 	}
 
 	if (frac) {
 		struct lconv const *l = localeconv();
 		char *dp = l ? l->decimal_point : NULL;
+		int len;
 
 		if (!dp || !*dp)
 			dp = ".";
-		snprintf(buf, sizeof(buf), "%d%s%" PRIu64 "%s", dec, dp, frac, suffix);
+
+		len = snprintf(buf, sizeof(buf), "%d%s%02" PRIu64, dec, dp, frac);
+		if (len > 0 && (size_t) len < sizeof(buf)) {
+			/* remove potential extraneous zero */
+			if (buf[len - 1] == '0')
+				buf[len--] = '\0';
+			/* append suffix */
+			xstrncpy(buf+len, suffix, sizeof(buf) - len);
+		} else
+			*buf = '\0';	/* snprintf error */
 	} else
 		snprintf(buf, sizeof(buf), "%d%s", dec, suffix);
 
@@ -706,11 +734,16 @@ int string_add_to_idarray(const char *list, int ary[], size_t arysz,
  * as a position in the 'ary' bit array. It means that the 'id' has to be in
  * range <0..N> where N < sizeof(ary) * NBBY.
  *
+ * If allow_range is enabled:
+ * An item ending in '+' also sets all bits in <0..N>.
+ * An item beginning with '+' also sets all bits in <N..allow_minus>.
+ *
  * Returns: 0 on success, <0 on error.
  */
 int string_to_bitarray(const char *list,
 		     char *ary,
-		     int (*name2bit)(const char *, size_t))
+		     int (*name2bit)(const char *, size_t),
+		     size_t allow_range)
 {
 	const char *begin = NULL, *p;
 
@@ -719,7 +752,7 @@ int string_to_bitarray(const char *list,
 
 	for (p = list; p && *p; p++) {
 		const char *end = NULL;
-		int bit;
+		int bit, set_lower = 0, set_higher = 0;
 
 		if (!begin)
 			begin = p;		/* begin of the level name */
@@ -731,11 +764,26 @@ int string_to_bitarray(const char *list,
 			continue;
 		if (end <= begin)
 			return -1;
+		if (allow_range) {
+			if (*(end - 1) == '+') {
+				end--;
+				set_lower = 1;
+			} else if (*begin == '+') {
+				begin++;
+				set_higher = 1;
+			}
+		}
 
 		bit = name2bit(begin, end - begin);
 		if (bit < 0)
 			return bit;
 		setbit(ary, bit);
+		if (set_lower)
+			while (--bit >= 0)
+				setbit(ary, bit);
+		else if (set_higher)
+			while (++bit < (int) allow_range)
+				setbit(ary, bit);
 		begin = NULL;
 		if (end && !*end)
 			break;
@@ -881,7 +929,8 @@ int streq_paths(const char *a, const char *b)
 	return 0;
 }
 
-char *strnappend(const char *s, const char *suffix, size_t b)
+/* concatenate two strings to a new string, the size of the second string is limited by @b */
+char *strnconcat(const char *s, const char *suffix, size_t b)
 {
         size_t a;
         char *r;
@@ -911,12 +960,14 @@ char *strnappend(const char *s, const char *suffix, size_t b)
         return r;
 }
 
-char *strappend(const char *s, const char *suffix)
+/* concatenate two strings to a new string */
+char *strconcat(const char *s, const char *suffix)
 {
-        return strnappend(s, suffix, suffix ? strlen(suffix) : 0);
+        return strnconcat(s, suffix, suffix ? strlen(suffix) : 0);
 }
 
-char *strfappend(const char *s, const char *format, ...)
+/* concatenate @s and string defined by @format to a new string */
+char *strfconcat(const char *s, const char *format, ...)
 {
 	va_list ap;
 	char *val, *res;
@@ -929,11 +980,36 @@ char *strfappend(const char *s, const char *format, ...)
 	if (sz < 0)
 		return NULL;
 
-	res = strnappend(s, val, sz);
+	res = strnconcat(s, val, sz);
 	free(val);
 	return res;
 }
 
+int strappend(char **a, const char *b)
+{
+	size_t al, bl;
+	char *tmp;
+
+	if (!a)
+		return -EINVAL;
+	if (!b || !*b)
+		return 0;
+	if (!*a) {
+		*a = strdup(b);
+		return !*a ? -ENOMEM : 0;
+	}
+
+	al = strlen(*a);
+	bl = strlen(b);
+
+	tmp = realloc(*a, al + bl + 1);
+	if (!tmp)
+		return -ENOMEM;
+	*a = tmp;
+	memcpy((*a) + al, b, bl + 1);
+	return 0;
+}
+
 static size_t strcspn_escaped(const char *s, const char *reject)
 {
         int escaped = 0;
@@ -952,6 +1028,37 @@ static size_t strcspn_escaped(const char *s, const char *reject)
         return n - escaped;
 }
 
+/*
+ * Like strchr() but ignores @c if escaped by '\', '\\' is interpreted like '\'.
+ *
+ * For example for @c='X':
+ *
+ *      "abcdXefgXh"    --> "XefgXh"
+ *	"abcd\XefgXh"   --> "Xh"
+ *	"abcd\\XefgXh"  --> "XefgXh"
+ *	"abcd\\\XefgXh" --> "Xh"
+ *	"abcd\Xefg\Xh"  --> (null)
+ *
+ *	"abcd\\XefgXh"  --> "\XefgXh"   for @c='\\'
+ */
+char *ul_strchr_escaped(const char *s, int c)
+{
+	char *p;
+	int esc = 0;
+
+	for (p = (char *) s; p && *p; p++) {
+		if (!esc && *p == '\\') {
+			esc = 1;
+			continue;
+		}
+		if (*p == c && (!esc || c == '\\'))
+			return p;
+		esc = 0;
+	}
+
+	return NULL;
+}
+
 /* Split a string into words. */
 const char *split(const char **state, size_t *l, const char *separator, int quoted)
 {
@@ -1010,12 +1117,145 @@ int skip_fline(FILE *fp)
 	} while (1);
 }
 
+
+/* compare two strings, but ignoring non-alnum and case of the characters, for example
+ * "Hello (123)!" is the same as "hello123".
+ */
+int ul_stralnumcmp(const char *p1, const char *p2)
+{
+	const unsigned char *s1 = (const unsigned char *) p1;
+	const unsigned char *s2 = (const unsigned char *) p2;
+	unsigned char c1, c2;
+
+	do {
+		do {
+			c1 = (unsigned char) *s1++;
+		} while (c1 != '\0' && !isalnum((unsigned int) c1));
+
+		do {
+			c2 = (unsigned char) *s2++;
+		} while (c2 != '\0' && !isalnum((unsigned int) c2));
+
+		if (c1 != '\0')
+			c1 = tolower(c1);
+		if (c2 != '\0')
+			c2 = tolower(c2);
+		if (c1 == '\0')
+			return c1 - c2;
+	} while (c1 == c2);
+
+	return c1 - c2;
+}
+
+/*
+ * Parses the first option from @optstr. The @optstr pointer is set to the beginning
+ * of the next option. The options string looks like 'aaa,bbb=data,foo,bar="xxx"'.
+ *
+ * Note this function is used by libmount to parse mount options. Be careful when modify.
+ *
+ * Returns -EINVAL on parse error, 1 at the end of optstr and 0 on success.
+ */
+int ul_optstr_next(char **optstr, char **name, size_t *namesz,
+		   char **value, size_t *valsz)
+{
+	int open_quote = 0;
+	char *start = NULL, *stop = NULL, *p, *sep = NULL;
+	char *optstr0;
+
+	assert(optstr);
+	assert(*optstr);
+
+	optstr0 = *optstr;
+
+	if (name)
+		*name = NULL;
+	if (namesz)
+		*namesz = 0;
+	if (value)
+		*value = NULL;
+	if (valsz)
+		*valsz = 0;
+
+	/* trim leading commas as to not invalidate option
+	 * strings with multiple consecutive commas */
+	while (optstr0 && *optstr0 == ',')
+		optstr0++;
+
+	for (p = optstr0; p && *p; p++) {
+		if (!start)
+			start = p;		/* beginning of the option item */
+		if (*p == '"')
+			open_quote ^= 1;	/* reverse the status */
+		if (open_quote)
+			continue;		/* still in quoted block */
+		if (!sep && p > start && *p == '=')
+			sep = p;		/* name and value separator */
+		if (*p == ',')
+			stop = p;		/* terminate the option item */
+		else if (*(p + 1) == '\0')
+			stop = p + 1;		/* end of optstr */
+		if (!start || !stop)
+			continue;
+		if (stop <= start)
+			return -EINVAL;
+
+		if (name)
+			*name = start;
+		if (namesz)
+			*namesz = sep ? sep - start : stop - start;
+		*optstr = *stop ? stop + 1 : stop;
+
+		if (sep) {
+			if (value)
+				*value = sep + 1;
+			if (valsz)
+				*valsz = stop - sep - 1;
+		}
+		return 0;
+	}
+
+	return 1;				/* end of optstr */
+}
+
 #ifdef TEST_PROGRAM_STRUTILS
 
+#include "cctype.h"
+
+struct testS {
+	char *name;
+	char *value;
+};
+
+static int test_strdup_to_member(int argc, char *argv[])
+{
+	struct testS *xx;
+
+	if (argc < 3)
+		return EXIT_FAILURE;
+
+	xx = calloc(1, sizeof(*xx));
+	if (!xx)
+		err(EXIT_FAILURE, "calloc() failed");
+
+	strdup_to_struct_member(xx, name, argv[1]);
+	strdup_to_struct_member(xx, value, argv[2]);
+
+	if (strcmp(xx->name, argv[1]) != 0 &&
+	    strcmp(xx->value, argv[2]) != 0)
+		errx(EXIT_FAILURE, "strdup_to_struct_member() failed");
+
+	printf("1: '%s', 2: '%s'\n", xx->name, xx->value);
+
+	free(xx->name);
+	free(xx->value);
+	free(xx);
+	return EXIT_SUCCESS;
+}
+
 static int test_strutils_sizes(int argc, char *argv[])
 {
 	uintmax_t size = 0;
-	char *hum, *hum2;
+	char *hum1, *hum2, *hum3;
 
 	if (argc < 2)
 		return EXIT_FAILURE;
@@ -1023,13 +1263,17 @@ static int test_strutils_sizes(int argc, char *argv[])
 	if (strtosize(argv[1], &size))
 		errx(EXIT_FAILURE, "invalid size '%s' value", argv[1]);
 
-	hum = size_to_human_string(SIZE_SUFFIX_1LETTER, size);
+	hum1 = size_to_human_string(SIZE_SUFFIX_1LETTER, size);
 	hum2 = size_to_human_string(SIZE_SUFFIX_3LETTER |
 				    SIZE_SUFFIX_SPACE, size);
+	hum3 = size_to_human_string(SIZE_SUFFIX_3LETTER |
+				    SIZE_SUFFIX_SPACE |
+				    SIZE_DECIMAL_2DIGITS, size);
 
-	printf("%25s : %20ju : %8s : %12s\n", argv[1], size, hum, hum2);
-	free(hum);
+	printf("%25s : %20ju : %8s : %12s : %13s\n", argv[1], size, hum1, hum2, hum3);
+	free(hum1);
 	free(hum2);
+	free(hum3);
 
 	return EXIT_SUCCESS;
 }
@@ -1045,17 +1289,111 @@ static int test_strutils_cmp_paths(int argc, char *argv[])
 	return EXIT_SUCCESS;
 }
 
+static int test_strutils_normalize(int argc, char *argv[])
+{
+	unsigned char *src, *dst, *org;
+	size_t sz, len;
+
+	if (argc < 2)
+		return EXIT_FAILURE;
+
+	org = (unsigned char *) strdup(argv[1]);
+	src = (unsigned char *) strdup((char *) org);
+	len = strlen((char *) src);
+	dst = malloc(len + 1);
+
+	if (!org || !src || !dst)
+		goto done;
+
+	/* two buffers */
+	sz = __normalize_whitespace(src, len, dst, len + 1);
+	printf("1: '%s' --> '%s' [sz=%zu]\n", src, dst, sz);
+
+	/* one buffer */
+	sz = normalize_whitespace(src);
+	printf("2: '%s' --> '%s' [sz=%zu]\n", org, src, sz);
+
+done:
+	free(src);
+	free(dst);
+	free(org);
+
+	return EXIT_SUCCESS;
+}
+
+static int test_strutils_cstrcasecmp(int argc, char *argv[])
+{
+	char *a, *b;
+
+	if (argc < 3)
+		return EXIT_FAILURE;
+
+	a = argv[1];
+	b = argv[2];
+
+	if (!a || !b)
+		return EXIT_FAILURE;
+
+	printf("cmp    '%s' '%s' = %d\n", a, b, strcasecmp(a, b));
+	printf("c_cmp  '%s' '%s' = %d\n", a, b, c_strcasecmp(a, b));
+	printf("c_ncmp '%s' '%s' = %d\n", a, b, c_strncasecmp(a, b, strlen(a)));
+
+	return EXIT_SUCCESS;
+}
+
 int main(int argc, char *argv[])
 {
-	if (argc == 3 && strcmp(argv[1], "--size") == 0)
+	if (argc == 3 && strcmp(argv[1], "--size") == 0) {
 		return test_strutils_sizes(argc - 1, argv + 1);
 
-	else if (argc == 4 && strcmp(argv[1], "--cmp-paths") == 0)
+	} else if (argc == 4 && strcmp(argv[1], "--cmp-paths") == 0) {
 		return test_strutils_cmp_paths(argc - 1, argv + 1);
 
-	else {
+	} else if (argc == 4 && strcmp(argv[1], "--strdup-member") == 0) {
+		return test_strdup_to_member(argc - 1, argv + 1);
+
+	} else if  (argc == 4 && strcmp(argv[1], "--stralnumcmp") == 0) {
+		printf("%s\n", ul_stralnumcmp(argv[2], argv[3]) == 0 ?
+				"match" : "dismatch");
+		return EXIT_SUCCESS;
+
+	} else if (argc == 4 && strcmp(argv[1], "--cstrcasecmp") == 0) {
+		return test_strutils_cstrcasecmp(argc - 1, argv + 1);
+
+	} else if (argc == 3 && strcmp(argv[1], "--normalize") == 0) {
+		return test_strutils_normalize(argc - 1, argv + 1);
+
+	} else if (argc == 3 && strcmp(argv[1], "--strtos64") == 0) {
+		printf("'%s'-->%jd\n", argv[2], strtos64_or_err(argv[2], "strtos64 failed"));
+		return EXIT_SUCCESS;
+	} else if (argc == 3 && strcmp(argv[1], "--strtou64") == 0) {
+		printf("'%s'-->%ju\n", argv[2], strtou64_or_err(argv[2], "strtou64 failed"));
+		return EXIT_SUCCESS;
+	} else if (argc == 3 && strcmp(argv[1], "--strtos32") == 0) {
+		printf("'%s'-->%d\n", argv[2], strtos32_or_err(argv[2], "strtos32 failed"));
+		return EXIT_SUCCESS;
+	} else if (argc == 3 && strcmp(argv[1], "--strtou32") == 0) {
+		printf("'%s'-->%u\n", argv[2], strtou32_or_err(argv[2], "strtou32 failed"));
+		return EXIT_SUCCESS;
+	} else if (argc == 3 && strcmp(argv[1], "--strtos16") == 0) {
+		printf("'%s'-->%hd\n", argv[2], strtos16_or_err(argv[2], "strtos16 failed"));
+		return EXIT_SUCCESS;
+	} else if (argc == 3 && strcmp(argv[1], "--strtou16") == 0) {
+		printf("'%s'-->%hu\n", argv[2], strtou16_or_err(argv[2], "strtou16 failed"));
+		return EXIT_SUCCESS;
+
+	} else if (argc == 4 && strcmp(argv[1], "--strchr-escaped") == 0) {
+		printf("\"%s\" --> \"%s\"\n", argv[2], ul_strchr_escaped(argv[2], *argv[3]));
+		return EXIT_SUCCESS;
+
+	} else {
 		fprintf(stderr, "usage: %1$s --size <number>[suffix]\n"
-				"       %1$s --cmp-paths <path> <path>\n",
+				"       %1$s --cmp-paths <path> <path>\n"
+				"       %1$s --strdup-member <str> <str>\n"
+				"       %1$s --stralnumcmp <str> <str>\n"
+				"       %1$s --cstrcasecmp <str> <str>\n"
+				"       %1$s --normalize <str>\n"
+				"       %1$s --strto{s,u}{16,32,64} <str>\n",
 				argv[0]);
 		exit(EXIT_FAILURE);
 	}
