@@ -14,6 +14,7 @@
 # GNU General Public License for more details.
 #
 
+TS_EXIT_NOTSUPP=2
 
 function ts_abspath {
 	cd $1
@@ -45,6 +46,16 @@ function ts_cd {
 	fi
 }
 
+function ts_separator {
+	local header="$1"
+	echo >> $TS_OUTPUT
+	if [ -z "$header" ]; then
+		echo "============================================" >> $TS_OUTPUT
+	else
+		echo "=====$header================================" >> $TS_OUTPUT
+	fi
+}
+
 function ts_report {
 	local desc=
 
@@ -65,13 +76,33 @@ function ts_report {
 }
 
 function ts_check_test_command {
-	if [ ! -x "$1" ]; then
-		ts_skip "${1##*/} not found"
-	fi
+	case "$1" in
+	"")
+		ts_failed "invalid test_command requested"
+		;;
+	*/*)
+		# paths
+		if [ ! -x "$1" ]; then
+			ts_skip "${1##*/} not found"
+		fi
+		;;
+	*)
+		# just command names (e.g. --use-system-commands)
+		local cmd=$1
+		type "$cmd" >/dev/null 2>&1
+	        if [ $? -ne 0 ]; then
+			if [ "$TS_NOSKIP_COMMANDS" = "yes" ]; then
+				ts_failed "missing in PATH: $cmd"
+			fi
+			ts_skip "missing in PATH: $cmd"
+		fi
+		;;
+	esac
 }
 
 function ts_check_prog {
 	local cmd=$1
+	[ -z "$cmd" ] && ts_failed "invalid prog requested"
 	type "$cmd" >/dev/null 2>&1 || ts_skip "missing in PATH: $cmd"
 }
 
@@ -91,12 +122,36 @@ function ts_check_losetup {
 	ts_skip "no loop-device support"
 }
 
-function ts_skip_subtest {
+function ts_check_wcsspn {
+	# https://gitlab.com/qemu-project/qemu/-/issues/1248
+	if [ -e "$TS_HELPER_SYSINFO" ] &&
+		[ "$("$TS_HELPER_SYSINFO" wcsspn-ok)" = "0" ]; then
+
+		ts_skip "non-functional widestring functions"
+	fi
+}
+
+function ts_check_native_byteorder {
+	if [ "$QEMU_USER" == "1" ] && [ ! -e /sys/kernel/cpu_byteorder ]; then
+		ts_skip "non-native byteorder"
+	fi
+}
+
+function ts_check_enotty {
+	# https://lore.kernel.org/qemu-devel/20230426070659.80649-1-thomas@t-8ch.de/
+	if [ -e "$TS_HELPER_SYSINFO" ] &&
+		[ "$("$TS_HELPER_SYSINFO" enotty-ok)" = "0" ]; then
+
+		ts_skip "broken ENOTTY return"
+	fi
+}
+
+function ts_report_skip {
 	ts_report " SKIPPED ($1)"
 }
 
 function ts_skip {
-	ts_skip_subtest "$1"
+	ts_report_skip "$1"
 
 	ts_cleanup_on_exit
 	exit 0
@@ -108,6 +163,12 @@ function ts_skip_nonroot {
 	fi
 }
 
+function ts_skip_qemu_user {
+	if [ "$QEMU_USER" == "1" ]; then
+		ts_skip "running under qemu-user emulation"
+	fi
+}
+
 function ts_failed_subtest {
 	local msg="FAILED"
 	local ret=1
@@ -130,7 +191,7 @@ function ts_failed {
 	exit $?
 }
 
-function ts_ok_subtest {
+function ts_report_ok {
 	if [ x"$1" == x"" ]; then
 		ts_report " OK"
 	else
@@ -139,7 +200,7 @@ function ts_ok_subtest {
 }
 
 function ts_ok {
-	ts_ok_subtest "$1"
+	ts_report_ok "$1"
 	exit 0
 }
 
@@ -148,6 +209,17 @@ function ts_log {
 	[ "$TS_VERBOSE" == "yes" ] && echo "$1"
 }
 
+function ts_logerr {
+	echo "$1" >> $TS_ERRLOG
+	[ "$TS_VERBOSE" == "yes" ] && echo "$1"
+}
+
+function ts_log_both {
+	echo "$1" >> $TS_OUTPUT
+	echo "$1" >> $TS_ERRLOG
+	[ "$TS_VERBOSE" == "yes" ] && echo "$1"
+}
+
 function ts_has_option {
 	NAME="$1"
 	ALL="$2"
@@ -191,24 +263,30 @@ function ts_init_core_env {
 	TS_SUBNAME=""
 	TS_NS="$TS_COMPONENT/$TS_TESTNAME"
 	TS_OUTPUT="$TS_OUTDIR/$TS_TESTNAME"
+	TS_ERRLOG="$TS_OUTDIR/$TS_TESTNAME.err"
 	TS_VGDUMP="$TS_OUTDIR/$TS_TESTNAME.vgdump"
+	TS_EXIT_CODE="$TS_OUTDIR/$TS_TESTNAME.exit_code"
 	TS_DIFF="$TS_DIFFDIR/$TS_TESTNAME"
 	TS_EXPECTED="$TS_TOPDIR/expected/$TS_NS"
+	TS_EXPECTED_ERR="$TS_TOPDIR/expected/$TS_NS.err"
 	TS_MOUNTPOINT="$TS_OUTDIR/${TS_TESTNAME}-mnt"
 }
 
 function ts_init_core_subtest_env {
 	TS_NS="$TS_COMPONENT/$TS_TESTNAME-$TS_SUBNAME"
 	TS_OUTPUT="$TS_OUTDIR/$TS_TESTNAME-$TS_SUBNAME"
+	TS_ERRLOG="$TS_OUTDIR/$TS_TESTNAME-$TS_SUBNAME.err"
 	TS_VGDUMP="$TS_OUTDIR/$TS_TESTNAME-$TS_SUBNAME.vgdump"
+	TS_EXIT_CODE="$TS_OUTDIR/$TS_TESTNAME-$TS_SUBNAME.exit_code"
 	TS_DIFF="$TS_DIFFDIR/$TS_TESTNAME-$TS_SUBNAME"
 	TS_EXPECTED="$TS_TOPDIR/expected/$TS_NS"
+	TS_EXPECTED_ERR="$TS_TOPDIR/expected/$TS_NS.err"
 	TS_MOUNTPOINT="$TS_OUTDIR/${TS_TESTNAME}-${TS_SUBNAME}-mnt"
 
-	rm -f $TS_OUTPUT $TS_VGDUMP
+	rm -f $TS_OUTPUT $TS_ERRLOG $TS_VGDUMP $TS_EXIT_CODE
 	[ -d "$TS_OUTDIR" ]  || mkdir -p "$TS_OUTDIR"
 
-	touch $TS_OUTPUT
+	touch $TS_OUTPUT $TS_ERRLOG $TS_EXIT_CODE
 	[ -n "$TS_VALGRIND_CMD" ] && touch $TS_VGDUMP
 }
 
@@ -220,8 +298,10 @@ function ts_init_env {
 	LANGUAGE="POSIX"
 	LC_ALL="POSIX"
 	CHARSET="UTF-8"
+	ASAN_OPTIONS="detect_leaks=0"
+	UBSAN_OPTIONS="print_stacktrace=1:print_summary=1:halt_on_error=1"
 
-	export LANG LANGUAGE LC_ALL CHARSET
+	export LANG LANGUAGE LC_ALL CHARSET ASAN_OPTIONS UBSAN_OPTIONS
 
 	mydir=$(ts_canonicalize "$mydir")
 
@@ -243,8 +323,24 @@ function ts_init_env {
 	top_srcdir=$(ts_abspath $top_srcdir)
 	top_builddir=$(ts_abspath $top_builddir)
 
-	# some ul commands search other ul commands in $PATH
-	export PATH="$top_builddir:$PATH"
+        if [ -e "$top_builddir/meson.conf" ]; then
+            . "$top_builddir/meson.conf"
+        fi
+
+	# We use helpser always from build tree
+	ts_helpersdir="${top_builddir}/"
+
+	TS_USE_SYSTEM_COMMANDS=$(ts_has_option "use-system-commands" "$*")
+	if [ "$TS_USE_SYSTEM_COMMANDS" == "yes" ]; then
+		# Don't define anything, just follow current PATH
+		ts_commandsdir=""
+	else
+		# The default is to use commands from build tree
+		ts_commandsdir="${top_builddir}/"
+
+		# some ul commands search other ul commands in $PATH
+		export PATH="$ts_commandsdir:$PATH"
+	fi
 
 	TS_SCRIPT="$mydir/$(basename $0)"
 	TS_SUBDIR=$(dirname $TS_SCRIPT)
@@ -260,19 +356,35 @@ function ts_init_env {
 	TS_OUTDIR="$top_builddir/tests/output/$TS_COMPONENT"
 	TS_DIFFDIR="$top_builddir/tests/diff/$TS_COMPONENT"
 
+	TS_NOLOCKS=$(ts_has_option "nolocks" "$*")
+	TS_LOCKDIR="$top_builddir/tests/output"
+
+	# Don't lock if flock(1) is missing
+	type "flock" >/dev/null 2>&1 || TS_NOLOCKS="yes"
+
 	ts_init_core_env
 
+	TS_NOSKIP_COMMANDS=$(ts_has_option "noskip-commands" "$*")
 	TS_VERBOSE=$(ts_has_option "verbose" "$*")
+	TS_SHOWDIFF=$(ts_has_option "show-diff" "$*")
 	TS_PARALLEL=$(ts_has_option "parallel" "$*")
 	TS_KNOWN_FAIL=$(ts_has_option "known-fail" "$*")
 	TS_SKIP_LOOPDEVS=$(ts_has_option "skip-loopdevs" "$*")
 	TS_PARSABLE=$(ts_has_option "parsable" "$*")
 	[ "$TS_PARSABLE" = "yes" ] || TS_PARSABLE="$TS_PARALLEL"
 
-	tmp=$( ts_has_option "memcheck" "$*")
+	tmp=$( ts_has_option "memcheck-valgrind" "$*")
 	if [ "$tmp" == "yes" -a -f /usr/bin/valgrind ]; then
 		TS_VALGRIND_CMD="/usr/bin/valgrind"
 	fi
+	tmp=$( ts_has_option "memcheck-asan" "$*")
+	if [ "$tmp" == "yes" ]; then
+		TS_ENABLE_ASAN="yes"
+	fi
+	tmp=$( ts_has_option "memcheck-ubsan" "$*")
+	if [ "$tmp" == "yes" ]; then
+		TS_ENABLE_UBSAN="yes"
+	fi
 
 	BLKID_FILE="$TS_OUTDIR/${TS_TESTNAME}.blkidtab"
 
@@ -280,6 +392,7 @@ function ts_init_env {
 	declare -a TS_SUID_USER
 	declare -a TS_SUID_GROUP
 	declare -a TS_LOOP_DEVS
+	declare -a TS_LOCKFILE_FD
 
 	if [ -f $TS_TOPDIR/commands.sh ]; then
 		. $TS_TOPDIR/commands.sh
@@ -287,15 +400,17 @@ function ts_init_env {
 
 	export BLKID_FILE
 
-	rm -f $TS_OUTPUT $TS_VGDUMP
+	rm -f $TS_OUTPUT $TS_ERRLOG $TS_VGDUMP $TS_EXIT_CODE
 	[ -d "$TS_OUTDIR" ]  || mkdir -p "$TS_OUTDIR"
 
-	touch $TS_OUTPUT
+	touch $TS_OUTPUT $TS_ERRLOG $TS_EXIT_CODE
 	[ -n "$TS_VALGRIND_CMD" ] && touch $TS_VGDUMP
 
 	if [ "$TS_VERBOSE" == "yes" ]; then
 		echo
 		echo "     script: $TS_SCRIPT"
+		echo "   commands: $ts_commandsdir"
+		echo "    helpers: $ts_helpersdir"
 		echo "    sub dir: $TS_SUBDIR"
 		echo "    top dir: $TS_TOPDIR"
 		echo "       self: $TS_SELF"
@@ -305,8 +420,10 @@ function ts_init_env {
 		echo "  namespace: $TS_NS"
 		echo "    verbose: $TS_VERBOSE"
 		echo "     output: $TS_OUTPUT"
+		echo "  error log: $TS_ERRLOG"
+		echo "  exit code: $TS_EXIT_CODE"
 		echo "   valgrind: $TS_VGDUMP"
-		echo "   expected: $TS_EXPECTED"
+		echo "   expected: $TS_EXPECTED{.err}"
 		echo " mountpoint: $TS_MOUNTPOINT"
 		echo
 	fi
@@ -347,61 +464,141 @@ function ts_init_suid {
 	TS_SUID_USER[$ct]=$(stat --printf="%U" $PROG)
 	TS_SUID_GROUP[$ct]=$(stat --printf="%G" $PROG)
 
-	chown root.root $PROG &> /dev/null
+	chown root:root $PROG &> /dev/null
 	chmod u+s $PROG &> /dev/null
 }
 
 function ts_init_py {
 	LIBNAME="$1"
 
-	[ -f "$top_builddir/py${LIBNAME}.la" ] || ts_skip "py${LIBNAME} not compiled"
+	if [ -f "$top_builddir/py${LIBNAME}.la" ]; then
+            # autotoolz build
+	    export LD_LIBRARY_PATH="$top_builddir/.libs:$LD_LIBRARY_PATH"
+	    export PYTHONPATH="$top_builddir/$LIBNAME/python:$top_builddir/.libs:$PYTHONPATH"
 
-	export LD_LIBRARY_PATH="$top_builddir/.libs:$LD_LIBRARY_PATH"
-	export PYTHONPATH="$top_builddir/$LIBNAME/python:$top_builddir/.libs:$PYTHONPATH"
+	    PYTHON_VERSION=$(awk '/^PYTHON_VERSION/ { print $3 }' $top_builddir/Makefile)
+	    PYTHON_MAJOR_VERSION=$(echo $PYTHON_VERSION | sed 's/\..*//')
 
-	export PYTHON_VERSION=$(awk '/^PYTHON_VERSION/ { print $3 }' $top_builddir/Makefile)
-	export PYTHON_MAJOR_VERSION=$(echo $PYTHON_VERSION | sed 's/\..*//')
+	    export PYTHON="python${PYTHON_MAJOR_VERSION}"
 
-	export PYTHON="python${PYTHON_MAJOR_VERSION}"
+        elif compgen -G "$top_builddir/$LIBNAME/python/py$LIBNAME*.so" >/dev/null; then
+            # mezon!
+            export PYTHONPATH="$top_builddir/$LIBNAME/python:$PYTHONPATH"
+
+        else
+            ts_skip "py${LIBNAME} not compiled"
+        fi
 }
 
-function ts_valgrind {
-	if [ -z "$TS_VALGRIND_CMD" ]; then
-		"$@"
-	else
-		$TS_VALGRIND_CMD --tool=memcheck --leak-check=full \
-				 --leak-resolution=high --num-callers=20 \
-				 --log-file="$TS_VGDUMP" "$@"
+function ts_run {
+	declare -a args
+	local asan_options="strict_string_checks=1:detect_stack_use_after_return=1:check_initialization_order=1:strict_init_order=1"
+
+	#
+	# ASAN mode
+	#
+	if [ "$TS_ENABLE_ASAN" == "yes" -o "$TS_ENABLE_UBSAN" == "yes" ]; then
+		args+=(env)
+		if [ "$TS_ENABLE_ASAN" == "yes" ]; then
+			# detect_leaks isn't supported on s390x: https://github.com/llvm/llvm-project/blob/master/compiler-rt/lib/lsan/lsan_common.h
+			if [ "$(uname -m)" != "s390x" ]; then
+				asan_options="$asan_options:detect_leaks=1"
+			fi
+			args+=(ASAN_OPTIONS=$asan_options)
+		fi
+		if [ "$TS_ENABLE_UBSAN" == "yes" ]; then
+			args+=(UBSAN_OPTIONS=print_stacktrace=1:print_summary=1:halt_on_error=1)
+		fi
 	fi
+
+	#
+	# valgrind mode
+	#
+	if [ -n "$TS_VALGRIND_CMD" ]; then
+		args+=(libtool --mode=execute "$TS_VALGRIND_CMD" --tool=memcheck --leak-check=full)
+		args+=(--leak-resolution=high --num-callers=20 --log-file="$TS_VGDUMP")
+	fi
+
+	"${args[@]}" "$@"
+	echo $? >$TS_EXIT_CODE
+}
+
+function ts_gen_diff_from {
+	local res=0
+	local expected="$1"
+	local output="$2"
+	local difffile="$3"
+
+	diff -u $expected $output > $difffile
+
+	if [ $? -ne 0 ] || [ -s $difffile ]; then
+		res=1
+		if [ "$TS_SHOWDIFF" == "yes" -a "$TS_KNOWN_FAIL" != "yes" ]; then
+			echo
+			echo "diff-{{{"
+			cat $difffile
+			echo "}}}-diff"
+			echo
+		fi
+	else
+		rm -f $difffile;
+	fi
+
+	return $res
 }
 
 function ts_gen_diff {
-	local res=0
+	local status_out=0
+	local status_err=0
+	local exit_code=0
 
 	[ -f "$TS_OUTPUT" ] || return 1
 	[ -f "$TS_EXPECTED" ] || TS_EXPECTED=/dev/null
 
 	# remove libtool lt- prefixes
 	sed --in-place 's/^lt\-\(.*\: \)/\1/g' $TS_OUTPUT
+	sed --in-place 's/^lt\-\(.*\: \)/\1/g' $TS_ERRLOG
 
 	[ -d "$TS_DIFFDIR" ] || mkdir -p "$TS_DIFFDIR"
-	diff -u $TS_EXPECTED $TS_OUTPUT > $TS_DIFF
 
-	if [ $? -ne 0 ] || [ -s $TS_DIFF ]; then
-		res=1
+	# error log is fully optional
+	[ -f "$TS_EXPECTED_ERR" ] || TS_EXPECTED_ERR=/dev/null
+	[ -f "$TS_ERRLOG" ] || TS_ERRLOG=/dev/null
+
+	if [ "$TS_COMPONENT" != "fuzzers" ]; then
+		ts_gen_diff_from $TS_EXPECTED $TS_OUTPUT $TS_DIFF
+		status_out=$?
+
+		ts_gen_diff_from $TS_EXPECTED_ERR $TS_ERRLOG $TS_DIFF.err
+		status_err=$?
 	else
-		rm -f $TS_DIFF;
+		# TS_EXIT_CODE is empty when tests aren't run with ts_run: https://github.com/util-linux/util-linux/issues/1072
+		# or when ts_finalize is called right after ts_finalize_subtest.
+		exit_code="$(cat $TS_EXIT_CODE)"
+		if [ -z "$exit_code" ]; then
+			exit_code=0
+		fi
+
+		if [ $exit_code -ne 0 ]; then
+			ts_gen_diff_from $TS_EXPECTED $TS_OUTPUT $TS_DIFF
+			ts_gen_diff_from $TS_EXPECTED_ERR $TS_ERRLOG $TS_DIFF.err
+		fi
 	fi
 
-	return $res
+	if [ $status_out -ne 0 -o $status_err -ne 0 -o $exit_code -ne 0 ]; then
+		return 1
+	fi
+	return 0
 }
 
 function tt_gen_mem_report {
-	[ -z "$TS_VALGRIND_CMD" ] && echo "$1"
-
-	grep -q -E 'ERROR SUMMARY: [1-9]' $TS_VGDUMP &> /dev/null
-	if [ $? -eq 0 ]; then
-		echo "mem-error detected!"
+	if [ -n "$TS_VALGRIND_CMD" ]; then
+		grep -q -E 'ERROR SUMMARY: [1-9]' $TS_VGDUMP &> /dev/null
+		if [ $? -eq 0 ]; then
+			echo "mem-error detected!"
+		fi
+	else
+		echo "$1"
 	fi
 }
 
@@ -413,7 +610,7 @@ function ts_finalize_subtest {
 		ts_failed_subtest "$1"
 		res=1
 	else
-		ts_ok_subtest "$(tt_gen_mem_report "$1")"
+		ts_report_ok "$(tt_gen_mem_report "$1")"
 	fi
 
 	[ $res -ne 0 ] && TS_NSUBFAILED=$(( $TS_NSUBFAILED + 1 ))
@@ -424,6 +621,13 @@ function ts_finalize_subtest {
 	return $res
 }
 
+function ts_skip_subtest {
+	ts_report_skip "$1"
+	# reset environment back to parental test
+	ts_init_core_env
+
+}
+
 function ts_finalize {
 	ts_cleanup_on_exit
 
@@ -449,18 +653,20 @@ function ts_cleanup_on_exit {
 	for idx in $(seq 0 $((${#TS_SUID_PROGS[*]} - 1))); do
 		PROG=${TS_SUID_PROGS[$idx]}
 		chmod a-s $PROG &> /dev/null
-		chown ${TS_SUID_USER[$idx]}.${TS_SUID_GROUP[$idx]} $PROG &> /dev/null
+		chown ${TS_SUID_USER[$idx]}:${TS_SUID_GROUP[$idx]} $PROG &> /dev/null
 	done
 
 	for dev in "${TS_LOOP_DEVS[@]}"; do
 		ts_device_deinit "$dev"
 	done
 	unset TS_LOOP_DEVS
+
+	ts_scsi_debug_rmmod
 }
 
 function ts_image_md5sum {
 	local img=${1:-"$TS_OUTDIR/${TS_TESTNAME}.img"}
-	echo $(md5sum "$img" | awk '{printf $1}') $(basename "$img")
+	echo $("$TS_HELPER_MD5" < "$img") $(basename "$img")
 }
 
 function ts_image_init {
@@ -483,7 +689,7 @@ function ts_device_init {
 	local dev
 
 	img=$(ts_image_init $1 $2)
-	dev=$($TS_CMD_LOSETUP --show -f "$img")
+	dev=$($TS_CMD_LOSETUP --show --partscan -f "$img")
 	if [ "$?" != "0" -o "$dev" = "" ]; then
 		ts_die "Cannot init device"
 	fi
@@ -502,16 +708,52 @@ function ts_device_deinit {
 	fi
 }
 
+function ts_blkidtag_by_devname()
+{
+	local tag=$1
+	local dev=$2
+	local out
+	local rval
+
+	out=$($TS_CMD_BLKID -p -s "$tag" -o value "$dev")
+	rval=$?
+	printf "%s\n" "$out"
+
+	test -n "$out" -a "$rval" = "0"
+	return $?
+}
+
 function ts_uuid_by_devname {
-	echo $($TS_CMD_BLKID -p -s UUID -o value $1)
+	ts_blkidtag_by_devname "UUID" "$1"
+	return $?
 }
 
 function ts_label_by_devname {
-	echo $($TS_CMD_BLKID -p -s LABEL -o value $1)
+	ts_blkidtag_by_devname "LABEL" "$1"
+	return $?
 }
 
 function ts_fstype_by_devname {
-	echo $($TS_CMD_BLKID -p -s TYPE -o value $1)
+	ts_blkidtag_by_devname "TYPE" "$1"
+	return $?
+}
+
+function ts_vfs_dump {
+	if [ "$TS_SHOWDIFF" == "yes" -a "$TS_KNOWN_FAIL" != "yes" ]; then
+		echo
+		echo "{{{{ VFS dump:"
+		findmnt
+		echo "}}}}"
+	fi
+}
+
+function ts_blk_dump {
+	if [ "$TS_SHOWDIFF" == "yes" -a "$TS_KNOWN_FAIL" != "yes" ]; then
+		echo
+		echo "{{{{ blkdevs dump:"
+		lsblk -o+FSTYPE
+		echo "}}}}"
+	fi
 }
 
 function ts_device_has {
@@ -519,25 +761,33 @@ function ts_device_has {
 	local VAL="$2"
 	local DEV="$3"
 	local vl=""
+	local res=""
 
-	case $TAG in
-		"TYPE") vl=$(ts_fstype_by_devname $DEV);;
-		"LABEL") vl=$(ts_label_by_devname $DEV);;
-		"UUID") vl=$(ts_uuid_by_devname $DEV);;
-		*) return 1;;
-	esac
+	vl=$(ts_blkidtag_by_devname "$TAG" "$DEV")
+	test $? = 0 -a "$vl" = "$VAL"
+	res=$?
 
-	if [ "$vl" == "$VAL" ]; then
-		return 0
+	if [ "$res" != 0 ]; then
+		ts_vfs_dump
+		ts_blk_dump
 	fi
-	return 1
+
+	return $res
 }
 
-function ts_device_has_uuid {
-	ts_uuid_by_devname "$1" | egrep -q '^[0-9a-z]{8}-[0-9a-z]{4}-[0-9a-z]{4}-[0-9a-z]{4}-[0-9a-z]{12}$'
+function ts_is_uuid()
+{
+	printf "%s\n" "$1" | grep -E -q '^[0-9a-z]{8}-[0-9a-z]{4}-[0-9a-z]{4}-[0-9a-z]{4}-[0-9a-z]{12}$'
 	return $?
 }
 
+function ts_udevadm_settle()
+{
+	local dev=$1 # optional, might be empty
+	shift        # all other args are tags, LABEL, UUID, ...
+	udevadm settle
+}
+
 function ts_mount {
 	local out
 	local result
@@ -565,10 +815,10 @@ function ts_mount {
 function ts_is_mounted {
 	local DEV=$(ts_canonicalize "$1")
 
-	grep -q $DEV /proc/mounts && return 0
+	grep -q "\(^\| \)$DEV " /proc/mounts && return 0
 
 	if [ "${DEV#/dev/loop/}" != "$DEV" ]; then
-		grep -q "/dev/loop${DEV#/dev/loop/}" /proc/mounts && return 0
+		grep -q "^/dev/loop${DEV#/dev/loop/} " /proc/mounts && return 0
 	fi
 	return 1
 }
@@ -579,6 +829,7 @@ function ts_fstab_open {
 
 function ts_fstab_close {
 	echo "# -->" >> /etc/fstab
+	sync /etc/fstab 2>/dev/null
 }
 
 function ts_fstab_addline {
@@ -590,13 +841,19 @@ function ts_fstab_addline {
 	echo "$SPEC   $MNT   $FS   $OPT   0   0" >> /etc/fstab
 }
 
+function ts_fstab_lock {
+	ts_lock "fstab"
+}
+
 function ts_fstab_add {
+	ts_fstab_lock
 	ts_fstab_open
 	ts_fstab_addline $*
 	ts_fstab_close
 }
 
 function ts_fstab_clean {
+	ts_have_lock "fstab" || return 0
 	sed --in-place "
 /# <!-- util-linux/!b
 :a
@@ -606,6 +863,9 @@ function ts_fstab_clean {
 }
 s/# <!-- util-linux.*-->//;
 /^$/d" /etc/fstab
+
+	sync /etc/fstab 2>/dev/null
+	ts_unlock "fstab"
 }
 
 function ts_fdisk_clean {
@@ -613,18 +873,103 @@ function ts_fdisk_clean {
 
 	# remove non comparable parts of fdisk output
 	if [ -n "${DEVNAME}" ]; then
-		sed -i -e "s@${DEVNAME}@<removed>@;" $TS_OUTPUT
+		# escape "@" with "\@" in $DEVNAME. This way sed correctly
+		# replaces paths containing "@" characters
+		sed -i -e "s@${DEVNAME//\@/\\\@}@<removed>@;" $TS_OUTPUT $TS_ERRLOG
 	fi
 
 	sed -i \
 		-e 's/Disk identifier:.*/Disk identifier: <removed>/' \
-		-e 's/Created a new.*/Created a new <removed>./' \
+		-e 's/Created a new partition.*/Created a new partition <removed>./' \
+		-e 's/Created a new .* disklabel .*/Created a new disklabel./' \
 		-e 's/^Device[[:blank:]]*Start/Device             Start/' \
 		-e 's/^Device[[:blank:]]*Boot/Device     Boot/' \
 		-e 's/Welcome to fdisk.*/Welcome to fdisk <removed>./' \
 		-e 's/typescript file.*/typescript file <removed>./' \
 		-e 's@^\(I/O size (minimum/op.* bytes /\) [1-9][0-9]* @\1 <removed> @' \
-		$TS_OUTPUT
+		$TS_OUTPUT $TS_ERRLOG
+}
+
+
+# https://stackoverflow.com/questions/41603787/how-to-find-next-available-file-descriptor-in-bash
+function ts_find_free_fd()
+{
+	local rco
+	local rci
+	for fd in {3..200}; do
+		rco="$(true 2>/dev/null >&${fd}; echo $?)"
+		rci="$(true 2>/dev/null <&${fd}; echo $?)"
+		if [[ "${rco}${rci}" = "11" ]]; then
+			echo "$fd"
+			return 0
+		fi
+	done
+	return 1
+}
+
+function ts_get_lock_fd {
+	local resource=$1
+	local fd
+
+	for fd in "${!TS_LOCKFILE_FD[@]}"; do
+		if [ "${TS_LOCKFILE_FD["$fd"]}" = "$resource" ]; then
+			echo "$fd"
+			return 0
+		fi
+	done
+	return 1
+}
+
+function ts_have_lock {
+	local resource=$1
+
+	test "$TS_NOLOCKS" = "yes" && return 0
+	ts_get_lock_fd "$resource" >/dev/null && return 0
+	return 1
+}
+
+function ts_lock {
+	local resource="$1"
+	local lockfile="${TS_LOCKDIR}/${resource}.lock"
+	local fd
+
+	if [ "$TS_NOLOCKS" == "yes" ]; then
+		return 0
+	fi
+
+	# Don't lock again
+	fd=$(ts_get_lock_fd "$resource")
+	if [ -n "$fd" ]; then
+		echo "[$$ $TS_TESTNAME] ${resource} already locked!"
+		return 0
+	fi
+
+	fd=$(ts_find_free_fd) || ts_skip "failed to find lock fd"
+
+	eval "exec $fd>$lockfile"
+	flock --exclusive "$fd" || ts_skip "failed to lock $resource"
+
+	TS_LOCKFILE_FD["$fd"]="$resource"
+	###echo "[$$ $TS_TESTNAME] Locked   $resource"
+}
+
+function ts_unlock {
+	local resource="$1"
+	local lockfile="${TS_LOCKDIR}/${resource}.lock"
+	local fd
+
+	if [ "$TS_NOLOCKS" == "yes" ]; then
+		return 0
+	fi
+
+	fd=$(ts_get_lock_fd "$resource")
+	if [ -n "$fd" ]; then
+		eval "exec $fd<&-"
+		TS_LOCKFILE_FD["$fd"]=""
+		###echo "[$$ $TS_TESTNAME] Unlocked $resource"
+	else
+		echo "[$$ $TS_TESTNAME] unlocking unlocked $resource!?"
+	fi
 }
 
 function ts_scsi_debug_init {
@@ -632,23 +977,25 @@ function ts_scsi_debug_init {
 	local t
 	TS_DEVICE="none"
 
+	ts_lock "scsi_debug"
+
 	# dry run is not really reliable, real modprobe may still fail
 	modprobe --dry-run --quiet scsi_debug &>/dev/null \
 		|| ts_skip "missing scsi_debug module (dry-run)"
 
 	# skip if still in use or removal of modules not supported at all
-	modprobe -r scsi_debug &>/dev/null \
-		|| ts_skip "cannot remove scsi_debug module (rmmod)"
-
-	# TODO validate that all devices are gone, add function ts_scsi_debug_rmmod
-	# to be used by the tests too. Tests which produce non-removable scsi
-	# devices should fail!
+	# We don't want a slow timeout here so we don't use ts_scsi_debug_rmmod!
+	modprobe -r scsi_debug &>/dev/null
+	if [ "$?" -eq 1 ]; then
+		ts_unlock "scsi_debug"
+		ts_skip "cannot remove scsi_debug module (rmmod)"
+	fi
 
 	modprobe -b scsi_debug "$@" &>/dev/null \
 		|| ts_skip "cannot load scsi_debug module (modprobe)"
 
 	# it might be still not loaded, modprobe.conf or whatever
-	lsmod | grep -q "^scsi_debug " \
+	lsmod 2>/dev/null | grep -q "^scsi_debug " \
 		|| ts_skip "scsi_debug module not loaded (lsmod)"
 
 	udevadm settle
@@ -656,15 +1003,58 @@ function ts_scsi_debug_init {
 	# wait for device if udevadm settle does not work
 	for t in 0 0.02 0.05 0.1 1; do
 		sleep $t
-		devname=$(grep --with-filename scsi_debug /sys/block/*/device/model) && break
+		devname=$(grep --no-messages --with-filename scsi_debug /sys/block/*/device/model) && break
 	done
-	[ -n "${devname}" ] || ts_die "timeout waiting for scsi_debug device"
+	[ -n "${devname}" ] || ts_skip "timeout waiting for scsi_debug device"
 
 	devname=$(echo $devname | awk -F '/' '{print $4}')
 	TS_DEVICE="/dev/${devname}"
 
 	# TODO validate that device is really up, for now just a warning on stderr
-	test -b $TS_DEVICE || echo "warning: scsi_debug device is still down"
+	test -b $TS_DEVICE || echo "warning: scsi_debug device is still down" >&2
+}
+
+# automatically called once in ts_cleanup_on_exit()
+function ts_scsi_debug_rmmod {
+	local err=1
+	local t
+	local lastmsg
+
+	# We must not run if we don't have the lock
+	ts_have_lock "scsi_debug" || return 0
+
+	# Return early most importantly in case we are not root or the module does
+	# not exist at all.
+	[ $UID -eq 0 ] || return 0
+	[ -n "$TS_DEVICE" ] || return 0
+	lsmod 2>/dev/null | grep -q "^scsi_debug " || return 0
+
+	udevadm settle
+
+	# wait for successful rmmod if udevadm settle does not work
+	for t in 0 0.02 0.05 0.1 1; do
+		sleep $t
+		lastmsg="$(modprobe -r scsi_debug 2>&1)" && err=0 && break
+	done
+
+	if [ "$err" = "1" ]; then
+		ts_log "rmmod failed: '$lastmsg'"
+		ts_log "timeout removing scsi_debug module (rmmod)"
+		return 1
+	fi
+	if lsmod | grep -q "^scsi_debug "; then
+		ts_log "BUG! scsi_debug still loaded"
+		return 1
+	fi
+
+	# TODO Do we need to validate that all devices are gone?
+	udevadm settle
+	test -b "$TS_DEVICE" && echo "warning: scsi_debug device is still up" >&2
+
+	# TODO unset TS_DEVICE, check that nobody uses it later, e.g. ts_fdisk_clean
+
+	ts_unlock "scsi_debug"
+	return 0
 }
 
 function ts_resolve_host {
@@ -719,3 +1109,51 @@ function ts_init_socket_to_file {
 		ts_skip "socat socket stopped listening"
 	fi
 }
+
+function ts_has_ncurses_support {
+	grep -q '#define HAVE_LIBNCURSES' ${top_builddir}/config.h
+	if [ $? == 0 ]; then
+		echo "yes"
+	else
+		echo "no"
+	fi
+}
+
+# Get path to the ASan runtime DSO the given binary was compiled with
+function ts_get_asan_rt_path {
+	local binary="${1?}"
+	local rt_path
+
+	ts_check_prog "ldd"
+	ts_check_prog "awk"
+
+	rt_path="$(ldd "$binary" | awk '/lib.+asan.*.so/ {print $3; exit}')"
+	if [ -n "$rt_path" -a -f "$rt_path" ]; then
+		echo "$rt_path"
+	fi
+}
+
+function ts_skip_exitcode_not_supported {
+	if [ $? -eq $TS_EXIT_NOTSUPP ]; then
+		ts_skip "functionality not implemented by system"
+	fi
+}
+
+function ts_inhibit_custom_colorscheme {
+	export XDG_CONFIG_HOME=/dev/null
+}
+
+function ts_is_virt {
+	type "systemd-detect-virt" >/dev/null 2>&1
+	if [ $? -ne 0 ]; then
+		return 1
+	fi
+
+	virt="$(systemd-detect-virt)"
+	for arg in "$@"; do
+		if [ "$virt" = "$arg" ]; then
+			return 0;
+		fi
+	done
+	return 1
+}
